/*
  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#include <stdlib.h>
#include <stdio.h>
#include <getopt.h>


#include "MatrixMultiplyBenchmark.h"

/* Options to be parsed */
static const char*         options = "m:n:p:t:c:h";
static const struct option long_options[] = {
  {"matrix-multiply-type",required_argument,NULL,'t'},
  {"cache-oblivious-base-size" ,required_argument,NULL,'c'}
};

/* Static functions */
static void usage();
static dmatrix * malloc_temporary_storage(const int nrows,const int ncols);
static void      free_temporary_storage(dmatrix* A);

/* Matrix Multiplication routines */
static void matrix_multiply0(double** C, double** A, double** B, const size_t m, const size_t n, const size_t p);
static void matrix_multiply1(double** C, double** A, double** B, const size_t m, const size_t n, const size_t p);
static void matrix_multiply2(double*  C, double*  A, double*  B, const size_t m, const size_t n, const size_t p);
static void matrix_multiply3(double** C, double** A, double** B, double** T    , const size_t m, const size_t n, const size_t p);
static void matrix_multiply4(double*  C, double*  A, double*  B, double*  T    , const size_t m, const size_t n, const size_t p);
static void matrix_multiply5(double*  C, double*  A, double*  B, double*  T    , const size_t m, const size_t n, const size_t p);
/* Read command-line parameters */
void read_input(int argc, char** argv, mmparameters* mmopt, dmatrix* A, dmatrix* B, dmatrix* C) {
  
  int opt = 0;
  int idx = 0;
  
  /* Set default values */
  A->nrows = DEFAULT_M;
  C->nrows = DEFAULT_M;
  A->ncols = DEFAULT_P;
  B->nrows = DEFAULT_P;
  B->ncols = DEFAULT_N;
  C->ncols = DEFAULT_N;
  
  mmopt->mmtype = DEFAULT_MMTYPE;
  mmopt->cache_oblivious_base_size  = DEFAULT_CACHE_OBLIVIOUS;
  /* Get options from command lines */
  for(;;) {
    
    opt = getopt_long(argc,argv,options,long_options,&idx);
    if (opt == -1) break;
    
    switch(opt) {
    case 'm':
      A->nrows = strtol(optarg,NULL,10);
      C->nrows = strtol(optarg,NULL,10);
      break;
    case 'p':
      A->ncols = strtol(optarg,NULL,10);
      B->nrows = strtol(optarg,NULL,10);
      break;
    case 'n':
      B->ncols = strtol(optarg,NULL,10);
      C->ncols = strtol(optarg,NULL,10);
      break;
    case 't':
      mmopt->mmtype = strtol(optarg,NULL,10);
      break;
    case 'c':
      mmopt->cache_oblivious_base_size = strtol(optarg,NULL,10);
      break;
    case 'h':
      usage();
      exit(EXIT_SUCCESS);      
      break;
    default:
      break;
    }

  }

}

/* Allocate a dmatrix if nrow and ncol are set */
void allocate_matrix(dmatrix* A) {
  /* Allocate enough raw space */
  A->values = (double**) malloc( A->nrows*A->ncols*sizeof(double) + A->nrows*sizeof(double*) );
  /* Initialize correctly the pointers */
  A->values[0] = (double*) &A->values[A->nrows];
  for (size_t ii = 1; ii < A->nrows; ++ii) {
    A->values[ii] = (double*) &A->values[ii-1][A->ncols];
  }
}

/* Initialize a matrix for benchmark */
void initialize_matrix(dmatrix* A) {  
  for( size_t ii = 0; ii < A->nrows; ++ii) {
      for( size_t jj = 0; jj < A->ncols; ++jj) {	
	if ( ii == jj) 
	  A->values[ii][jj] = 2.0;
	else if ( (ii - jj == 1) || (ii - jj == -1) )
	  A->values[ii][jj] = -1.0;
      }
  }
}

/* Benchmark test */
void benchmark_matrix_multiplication(dmatrix* C, dmatrix* A, dmatrix * B, const mmparameters * opt) {

  dmatrix * T;

  printf("\n");
  printf("Matrix A: %d x %d\n",A->nrows,A->ncols);
  printf("Matrix B: %d x %d\n",B->nrows,B->ncols);
  printf("Matrix C: %d x %d\n",C->nrows,C->ncols);
  printf("\n");
  
  switch(opt->mmtype) {
  case 0:
    matrix_multiply0(C->values,A->values,B->values,A->nrows,B->ncols,A->ncols);
    break;
  case 1:
    matrix_multiply1(C->values,A->values,B->values,A->nrows,B->ncols,A->ncols);
    break;

  case 2:    
    matrix_multiply2(C->values[0],A->values[0],B->values[0],A->nrows,B->ncols,A->ncols);
    break;
  case 3:
    T = malloc_temporary_storage(B->ncols,B->nrows);
    matrix_multiply3(C->values,A->values,B->values,T->values,A->nrows,B->ncols,A->ncols);
    free_temporary_storage(T);
    break;
  case 4:
    T = malloc_temporary_storage(B->ncols,B->nrows);
    matrix_multiply4(C->values[0],A->values[0],B->values[0],T->values[0],A->nrows,B->ncols,A->ncols);
    free_temporary_storage(T);
    break;
  case 5:
    T = malloc_temporary_storage(B->ncols,B->nrows);
    matrix_multiply5(C->values[0],A->values[0],B->values[0],T->values[0],A->nrows,B->ncols,A->ncols);
    free_temporary_storage(T);
    break;
  default:
    printf("Implementation number (-t %d) out of range",opt->mmtype);
    usage();
    free_matrix(A);
    free_matrix(B);
    free_matrix(C);
    exit(EXIT_FAILURE);
    break;
  }
}

/* Free memory */
void free_matrix(dmatrix* A) {
  if (A->values) free(A->values);
}

/* Print matrix for debug purpose */
void print_matrix(const dmatrix* A) {
  printf("\n");
  for( size_t ii = 0; ii < A->nrows; ++ii) {
    for( size_t jj = 0; jj < A->ncols; ++jj) {
      printf("%3.3f\t",A->values[ii][jj]);
    }
    printf("\n");
  }
  printf("\n");
}

static void usage() {
  printf("\n./Program to benchmark various forms of double-valued dense matrix multiplication\n");
  printf("\n./MatrixMultiply.x [options] \n");
  printf("\n");
  printf("C = A*B\n");
  printf("\n");
  printf("-m : rows of A and C (default: %d)\n",DEFAULT_M);
  printf("-n : columns of B and C (default: %d)\n",DEFAULT_N);
  printf("-p : columns of A and rows of B (default: %d)\n",DEFAULT_P);
  printf("-t : select the implementation to test (default: %d)\n",DEFAULT_MMTYPE);
  printf("\t0 => matrix-multiplication using double** pointers\n");
  printf("\t1 => matrix-multiplication using double** pointers and loop permutation\n");
  printf("\t2 => matrix-multiplication using double*  pointers and restrict qualifiers\n");
  printf("\t3 => matrix-multiplication using double** pointers and transposition of matrix B\n");
  printf("\t4 => matrix-multiplication using double*  pointers, restrict qualifiers and transposition of matrix B\n");
  printf("\t5 => matrix-multiplication using double*  pointers, restrict qualifiers, transposition of matrix B and cache-oblivious divide-and-conquer strategy\n");
  printf("-c : linear size to be used as a base-case threshold in cache-oblivious algorithms (default: %d)\n",DEFAULT_CACHE_OBLIVIOUS);
  printf("\n");
}

static dmatrix * malloc_temporary_storage(const int nrows,const int ncols) {
  dmatrix * A;
  A = malloc(sizeof(dmatrix));
  A->nrows = nrows;
  A->ncols = ncols;
  allocate_matrix(A);
  return A;
}

static void free_temporary_storage(dmatrix* A) {
  free_matrix(A);
  free(A);
}
static inline size_t min(const size_t x, const size_t y);
static inline size_t min(const size_t x, const size_t y) {
  return ( y ^ ((x ^ y) & -(x < y)) );
}



/* Simple matrix multiplication with double** */
static void matrix_multiply0(double** C, double** A, double** B, size_t m, size_t n, size_t p) {
  printf("Matrix-multiplication using double** pointers\n");
  /* Matrix multiplication */
  for( size_t ii = 0; ii < m; ++ii) {
    for( size_t jj = 0; jj < n; ++jj) {
      for( size_t kk = 0; kk < p; ++kk){ 
	C[ii][jj] += A[ii][kk]*B[kk][jj];
      }
    }
  }
}
/* Simple matrix multiplication with double* */
/* Optimization 1: loop permutation */
static void matrix_multiply1(double** C, double** A, double** B, size_t m, size_t n, size_t p) {
  printf("Matrix-multiplication using double** pointers\n");
  /* Matrix multiplication */
  for( size_t ii = 0; ii < m; ++ii) {
      for( size_t kk = 0; kk < p; ++kk){
        for( size_t jj = 0; jj < n; ++jj) {
        C[ii][jj] += A[ii][kk]*B[kk][jj];
      }
    }
  }

}



/* Simple matrix multiplication with double* */
/* Optimization 2: remove aliasing and simplify dereferentation */
static void matrix_multiply2(double*  restrict C, double*  restrict A, double*  restrict B, const size_t m, const size_t n, const size_t p) {
  printf("Matrix-multiplication using double*  pointers and restrict qualifiers\n");
  /* Matrix multiplication */

  for( size_t ii = 0; ii < m; ++ii) {
        for( size_t jj = 0; jj < n; ++jj) {
         for( size_t kk = 0; kk < p; ++kk){
        C[ii*n + jj] += A[ii*p + kk]*B[kk*n + jj];
      }
    }
  }
	
}

/* Optimization 3: 0 + transpose B */
static void matrix_multiply3(double** C, double** A, double** B, double** T, const size_t m, const size_t n, const size_t p) {
  printf("Matrix-multiplication using double** pointers and transposition of matrix B\n");
  /* Transpose B into T */
  /* Matrix multiplication */
   /* Transpose B into T */
for( size_t ii = 0; ii < n; ++ii)
    for( size_t jj = 0; jj < p; ++jj)
      T[ii][jj] = B[jj][ii];
  /* Matrix multiplication */
  for( size_t ii = 0; ii < m; ++ii) {
    for( size_t jj = 0; jj < n; ++jj) {
      for( size_t kk = 0; kk < p; ++kk){
        C[ii][jj] += A[ii][kk]*T[jj][kk];
      }
    }
  }

}  

/* Optimization 4: 2 + transpose B + restrict */
static void matrix_multiply4(double* restrict C, double* restrict A, double* restrict B, double * T, const size_t m, const size_t n, const size_t p) {
  printf("Matrix-multiplication using double*  pointers, restrict qualifiers and transposition of matrix B\n");
/* Transpose B into T */
  /* Matrix multiplication */
 /* Transpose B into T */
  for( size_t ii = 0; ii < n; ++ii)
    for( size_t jj = 0; jj < p; ++jj)
      T[ii*p + jj] = B[jj*n + ii];
 
  double temp;
  /* Matrix multiplication */
  for( size_t ii = 0; ii < m; ++ii) {
    for( size_t jj = 0; jj < n; ++jj) {
      for( size_t kk = 0; kk < p; ++kk){
        C[ii*n + jj] = A[ii*p + kk]*T[jj*p + kk];
      }
    }
  }


}

/* Optimization 5: 4 + cache blocking by maximum dimension */
static void matrix_multiply5(double* restrict C, double* restrict A, double* restrict B, double* restrict T, const size_t m, const size_t n, const size_t p) {
  printf("Matrix-multiplication using double*  pointers, restrict qualifiers, transposition of matrix B and cache-blocking  iterative strategy\n");
  /* Transpose B into T */
  for( size_t ii = 0; ii < n; ++ii)
    for( size_t jj = 0; jj < p; ++jj)
      T[ii*p + jj] = B[jj*n + ii];    
  /* Matrix multiplication */
  for( size_t ii = 0; ii < m; ii+= TILE_II) {
    for( size_t jj = 0; jj < n; jj += TILE_JJ) {
      for( size_t kk = 0; kk < p; kk += TILE_KK){ 
	for( size_t iix = ii; iix < min(ii+TILE_II,m); ++iix )
	  for( size_t jjx = jj; jjx < min(jj+TILE_JJ,n); ++jjx )
	    for( size_t kkx = kk; kkx < min(kk+TILE_KK,p); ++kkx )
	      C[iix*n + jjx] += A[iix*p + kkx]*T[jjx*p + kkx];
      }
    }
  }
}
